/*********************************************************/
/* BootLoader for ATMegaXXX & MegaLoad                   */
/* Version 1.1 April 2003                                */
/* 1.0 : First Release                                   */
/* 1.1 : Add auto baud rate, code smaller                */
/*                                                       */
/* Sylvain.Bissonnette@MicroSyl.com                      */
/*********************************************************/
/* Modified by Arthur de Beun for use with WINAVR        */
/* 2003 July 30                                          */

/* DISCLAIMER                                            */
/* Tested with ATmega8, 16                               */
/* Others not tested yet                                 */

/* in AVRStudio with STK500 tick the following:          */
/*   Boot Flash section size = 512 words ...             */
/*   Boot Reset vector Enabled ...                       */

/* *** SELECT THE APPROPRIATE MICRO IN THE MAKEFILE ***  */
/* **** Select CPU clock speed and baudrate below *****  */

/*********************************************************/
/*                  I N C L U D E S                      */
/*********************************************************/

#include <avr/interrupt.h>
#include <avr/io.h>
#include "main.h"
#include "assembly.h"

/*********************************************************/
/*                   D E F I N E S                       */
/*********************************************************/

//#define CPUCLK (8000000L)       // cpu crystal [Hz]
#define CPUCLK (11059200L)      // cpu crystal [Hz]
#define BAUDRATE (38400L)       // serial port baudrate
#define UBRR_INIT (CPUCLK/(16*BAUDRATE)-1)

#if defined (__AVR_ATmega8__)
#define  DeviceID    'A'        // Mega8
#define  FlashSize   'l'        // Flash 8k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'R'        // 64 Bytes
#define  PageByte    64         // 64 Bytes
#define  PageMul     6
#define  EXECCODE    0xffc0
#define  RXD_PULLUP  (PORTD |= 0x01)
#define  UCSRB_X     UCSRB
#define  UCSRA_X     UCSRA
#define  UCSRC_X     UCSRC
#define  UBRRH_X     UBRRH
#define  UBRRL_X     UBRRL
#define  UDR_X       UDR
#define  INTVECT     GICR

#elif defined (__AVR_ATmega16__)
#define  DeviceID    'B'        // Mega16
#define  FlashSize   'm'        // Flash 16k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'S'        // 128 Bytes
#define  PageByte    128        // 128 Bytes
#define  PageMul     7
#define  EXECCODE    0xff80
#define  RXD_PULLUP  (PORTD |= 0x01)
#define  UCSRB_X     UCSRB
#define  UCSRA_X     UCSRA
#define  UCSRC_X     UCSRC
#define  UBRRH_X     UBRRH
#define  UBRRL_X     UBRRL
#define  UDR_X       UDR
#define  INTVECT     GICR

#elif defined (__AVR_ATmega32__)
#define  DeviceID    'E'        // Mega32
#define  FlashSize   'n'        // Flash 32k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'S'        // 128 Bytes
#define  PageByte    128        // 128 Bytes
#define  PageMul     7
#define  EXECCODE    0xff80
#define  RXD_PULLUP  (PORTD |= 0x01)
#define  UCSRB_X     UCSRB
#define  UCSRA_X     UCSRA
#define  UCSRC_X     UCSRC
#define  UBRRH_X     UBRRH
#define  UBRRL_X     UBRRL
#define  UDR_X       UDR
#define  INTVECT     GICR

#elif defined (__AVR_ATmega64__)
#define  DeviceID    'C'        // Mega64
#define  FlashSize   'o'        // Flash 64k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'T'        // 256 Bytes
#define  PageByte    256        // 256 Bytes
#define  PageMul     8
#define  EXECCODE    0xff00
#if 0                           // select UART, 0 or 1
#define  RXD_PULLUP  (PORTD |= 0x04)
#define  UCSRB_X     UCSR1B
#define  UCSRA_X     UCSR1A
#define  UCSRC_X     UCSR1C
#define  UBRRH_X     UBRR1H
#define  UBRRL_X     UBRR1L
#define  UDR_X       UDR1
#else
#define  RXD_PULLUP  (PORTE |= 0x01)
#define  UCSRB_X     UCSR0B
#define  UCSRA_X     UCSR0A
#define  UCSRC_X     UCSR0C
#define  UBRRH_X     UBRR0H
#define  UBRRL_X     UBRR0L
#define  UDR_X       UDR0
#endif
#define  INTVECT     MCUCR

#elif defined (__AVR_ATmega128__)
#define  DeviceID    'D'        // Mega128
#define  FlashSize   'p'        // Flash 128k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'T'        // 256 Bytes
#define  PageByte    256        // 256 Bytes
#define  PageMul     8
#define  EXECCODE    0xffff
#if 0                           // select UART, 0 or 1
#define  RXD_PULLUP  (PORTD |= 0x04)
#define  UCSRB_X     UCSR1B
#define  UCSRA_X     UCSR1A
#define  UCSRC_X     UCSR1C
#define  UBRRH_X     UBRR1H
#define  UBRRL_X     UBRR1L
#define  UDR_X       UDR1
#else
#define  RXD_PULLUP  (PORTE |= 0x01)
#define  UCSRB_X     UCSR0B
#define  UCSRA_X     UCSR0A
#define  UCSRC_X     UCSR0C
#define  UBRRH_X     UBRR0H
#define  UBRRL_X     UBRR0L
#define  UDR_X       UDR0
#endif
#define  INTVECT     MCUCR

#elif defined (__AVR_ATmega162__)
#define  DeviceID    'F'        // Mega162
#define  FlashSize   'm'        // Flash 16k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'S'        // 128 Bytes
#define  PageByte    128        // 128 Bytes
#define  PageMul     7
#define  EXECCODE    0xff80
#if 0                           // select UART, 0 or 1
#define  RXD_PULLUP  (PORTB |= 0x04)
#define  UCSRB_X     UCSR1B
#define  UCSRA_X     UCSR1A
#define  UCSRC_X     UCSR1C
#define  UBRRH_X     UBRR1H
#define  UBRRL_X     UBRR1L
#define  UDR_X       UDR1
#else
#define  RXD_PULLUP  (PORTD |= 0x01)
#define  UCSRB_X     UCSR0B
#define  UCSRA_X     UCSR0A
#define  UCSRC_X     UCSR0C
#define  UBRRH_X     UBRR0H
#define  UBRRL_X     UBRR0L
#define  UDR_X       UDR0
#endif
#define  INTVECT     GICR

#elif defined (__AVR_ATmega169__)
#define  DeviceID    'B'        // Mega16
#define  FlashSize   'm'        // Flash 16k
#define  BootSize    'c'        // 512 Words
#define  PageSize    'S'        // 128 Bytes
#define  PageByte    128        // 128 Bytes
#define  PageMul     7
#define  EXECCODE    0xff80
#define  RXD_PULLUP  (PORTE |= 0x01)
#define  UCSRB_X     UCSRB
#define  UCSRA_X     UCSRA
#define  UCSRC_X     UCSRC
#define  UBRRH_X     UBRRH
#define  UBRRL_X     UBRRL
#define  UDR_X       UDR
#define  INTVECT     MCUCR

#else
#warning "device type not defined"
#endif

#define  FALSE       0
#define  TRUE        1

/*********************************************************/
/*           G L O B A L    V A R I A B L E S            */
/*********************************************************/

unsigned char PageBuffer[PageByte];
unsigned int PageAddress;

/*********************************************************/

void BootLoad(void)
{
  TxChar(DeviceID);
  TxChar(FlashSize);
  TxChar(BootSize);
  TxChar(PageSize);
  RxChar();

  TxChar('!');
  while(TRUE)
  {
    GetPageNumber();
    if (PageAddress == EXECCODE)
      ExecCode();
    if (GetPage())
    {
      WriteFlash();
      if (CheckFlash())
        TxChar('!');
      else
        TxChar('@');
    }
    else
      TxChar('@');
  }
}

/*********************************************************/

void GetPageNumber(void)
{
  while(!IsChar());
  PageAddress = RxChar() << 8;
#if defined (__AVR_ATmega128__)
  if (PageAddress)
    PageAddress16 = 1;
  else
    PageAddress16 = 0;
#endif
  while(!IsChar());
  PageAddress += RxChar();
#if defined (__AVR_ATmega128__)
  if (PageAddress != EXECCODE)  // if EXECCODE, don't shift
#endif
  PageAddress = PageAddress << PageMul;
} 

/*********************************************************/

void ExecCode(void)
{
#if defined (__AVR_ATmega128__)
  RAMPZ = 0;
#endif
  INTVECT = 0x01;               // Enable interrupt vector select
  INTVECT = 0x00;               // Move interrupt vector to flash
  asm volatile ("push r1" "\n\t"
                "push r1" "\n\t"
                "ret"     "\n\t"
                ::);
}

/*********************************************************/

char GetPage(void)
{
  unsigned int i;
  unsigned char LocalCheckSum = 0;
  unsigned char CheckSum = 0;

  for (i = 0; i < PageByte; i++)
  {
    while(!IsChar());
    PageBuffer[i] = RxChar();
    LocalCheckSum += PageBuffer[i];
  }

  while(!IsChar());
  CheckSum = RxChar();
  if (LocalCheckSum == CheckSum)
    return TRUE;
  else
    return FALSE;
}

/*********************************************************/

void WriteFlash(void)
{
  unsigned int i;
  unsigned int TempInt;

  for (i = 0; i < PageByte; i += 2)
  {
    TempInt = PageBuffer[i] + (PageBuffer[i+1] << 8);
    fill_temp_buffer(TempInt,i);
  }
#if defined (__AVR_ATmega128__)
  if (PageAddress16)
    RAMPZ = 1;
#endif
  write_page(PageAddress, 0x03); //Perform page ERASE
  write_page(PageAddress, 0x05); //Perform page write

  enableRWW();
}

/*********************************************************/

char CheckFlash(void)
{
  unsigned int i;
  unsigned int TempInt;
  unsigned int TempInt2;

  for (i = 0; i < PageByte; i += 2)
  {
#if defined (__AVR_ATmega128__)
	if (PageAddress16)
      RAMPZ = 1;
#endif
    TempInt = read_program_memory(PageAddress + i, 0x00);
#if defined (__AVR_ATmega128__)
	RAMPZ = 0;
#endif
    TempInt2 = PageBuffer[i] + (PageBuffer[i+1]<<8);
    if (TempInt != TempInt2)
      return FALSE;
  }
  return TRUE;
}

/*********************************************************/

unsigned char IsChar(void)
{
  if (UCSRA_X & 0x80)
    return TRUE;
  else
    return FALSE;
}

/*********************************************************/

unsigned char RxChar(void)
{
  return UDR_X;                 // Read char
}

/*********************************************************/

void TxChar(unsigned char ch)
{
  while (!(UCSRA_X & 0x20));    // Wait for empty transmit buffer
  UDR_X = ch;                   // Write char
}

void main(void)
{
  cli();                        // disable all interrupts
  RXD_PULLUP;                   // pull-up on RxD
  UCSRB_X = 0x00;               // disable while setting baud rate
  UCSRA_X = 0x00;
  UCSRC_X = 0x86;               // Asyn,NoParity,1StopBit,8Bit,
  UBRRH_X = 0x00;               // set baud rate hi
  UCSRB_X = 0x18;               // Rx enable Tx Enable
  UBRRL_X = UBRR_INIT;
  fv_wait();                    // need a delay for this to work properly on the STK500
  RxChar();
  TxChar('>');
  fv_wait();
  if (RxChar() == '<')
  {
    BootLoad();
    ExecCode();
  }
  ExecCode();
}

